<?php

namespace Drupal\couchbasedrupal;

/**
 * We need some extra magic around
 * the bucket.
 */
class CouchbaseBucket {

  /**
   * Couchbase bucket
   *
   * @var \CouchbaseBucket
   */
  protected $bucket;

  /**
   * Bucket name.
   *
   * @var string
   */
  protected $name;

  /**
   * Bucket password.
   *
   * @var string
   */
  protected $password;

  /**
   * If N1ql client is enabled.
   *
   * @var bool
   */
  protected $N1qlEnabled = FALSE;

  /**
   * Runs a couchbase operation with retry logic.
   *
   * @param callable $callable
   */
  protected function executeCouchbaseOperation($callable) {
    // Number of failures.
    $failures = 0;
    // Maximum retry time (s).
    $max_downtime = 10;
    // Operation start time.
    $start = microtime(TRUE);
    while (TRUE) {
      try {
        return call_user_func($callable);
      }
      catch (\Exception $e) {
        // If we have already failed, or this is not a
        // transient error code rethrow.
        if (!in_array((string) $e->getCode(), CouchbaseExceptionCodes::getTransientErrors())) {
          throw $e;
        }
        $failures++;
        // Limit the number of retries.
        if ((microtime(TRUE) - $start) > $max_downtime) {
          throw $e;
        }
        // Wait time proportional
        // to the number of failures
        usleep($failures * (0.5 * 1000000));
      }
    }
  }

  /**
   * Fix expiration for better precision.
   *
   * @see http://docs.couchbase.com/developer/dev-guide-3.0/doc-expiration.html
   */
  public function FixExpiration($expire) {
    // Because time() might not be exactly the same in PHP and in Couchbase,
    // use reltive TTL() instead of absolute for small timeouts.
    $timeout = $expire - time();
    if ($timeout < 30 * 24 * 60 * 60) {
      $result = (int) $timeout;
    }
    else {
      $result = (int) $expire;
    }

    return $result > 0 ? $result : 1;
  }

  /**
   * Get an instance of CouchbaseBucket
   *
   * @param \CouchbaseCluster $cluster
   * @param mixed $name
   * @param TranscoderInterface $transcoder
   */
  public function __construct(\CouchbaseCluster $cluster, $name, $password = NULL, TranscoderInterface $transcoder = NULL) {
    $args = [$name];
    if (!empty($password)) {
      $args[] = $password;
    }
    $this->executeCouchbaseOperation(function() use ($cluster, $args) {
      $this->bucket = call_user_func_array([$cluster, 'openBucket'], $args);
    });
    // If the user provides a transcoder, use it.
    if ($transcoder) {
      $this->bucket->setTranscoder([$transcoder, 'encode'], [$transcoder, 'decode']);
    }
    $this->name = $name;
    $this->password = $password;
  }

  /**
   * Retrieve an item from the bucket. Returns
   * FALSE if it does not exist.
   *
   * @param string $key
   * @return mixed|bool
   */
  public function get($key) {
    try {
      return $this->executeCouchbaseOperation(
        function() use ($key) {
          return $this->bucket->get($key);
        }
      );
    }
    catch (\Exception $e) {
      if ((string) $e->getCode() == CouchbaseExceptionCodes::KEY_DOES_NOT_EXIST) {
        return FALSE;
      }
      throw $e;
    }
  }

  /**
   * Retrieve a list of items from the bucket.
   *
   * @param string[] $key
   * @return array
   */
  public function getMultiple($keys) {
    return $this->executeCouchbaseOperation(
      function() use ($keys) {
        return $this->bucket->get($keys);
      }
    );
  }

  /**
   * Remove an item from the bucket without
   * throwing and exceptino if it does not exist.
   *
   * @param string $keys
   *   A key or an array of keys to remove.
   * @return bool
   *   Always returns TRUE.
   */
  public function remove($key) {
    try {
      // We can delete all of them at once!
      $this->executeCouchbaseOperation(
        function() use ($key) {
          $this->bucket->remove($key);
        }
      );
    }
    catch (\Exception $e) {
      if ((string) $e->getCode() == CouchbaseExceptionCodes::KEY_DOES_NOT_EXIST) {
        return FALSE;
      }
      throw $e;
    }
    return TRUE;
  }

  /**
   * Remove an item from the bucket without
   * throwing and exceptino if it does not exist.
   *
   * @param string[] $keys
   *   An array of keys to remove.
   * @return bool
   *   Always returns TRUE.
   */
  public function removeMultiple($keys) {
    // We can delete all of them at once!
    $this->executeCouchbaseOperation(
      function() use ($keys) {
        return $this->bucket->remove($keys);
      }
    );
    return TRUE;
  }

  /**
   * Upsert an intem in the bucket.
   *
   * @param string $id
   * @param mixed $val
   * @param array $options
   */
  public function upsert($id, $val = NULL, $options = array()) {
    return $this->executeCouchbaseOperation(
      function() use ($id, $val, $options) {
        return $this->bucket->upsert($id, $val, $options);
      }
    );
  }

  /**
   * Insert an item in a bucket. If the item
   * already exists returns FALSE.
   *
   * @param string $id
   * @param mixed $val
   * @param mixed $options
   * @return bool
   */
  public function insert($id, $val = NULL, $options = array()) {
    try {
      $this->executeCouchbaseOperation(
        function() use ($id, $val, $options) {
          $this->bucket->insert($id, $val, $options);
        }
      );
      return TRUE;
    }
    catch (\Exception $e) {
      // 12,13 are document already exists or document does not exist.
      if (in_array((string) $e->getCode(), [CouchbaseExceptionCodes::KEY_DOES_NOT_EXIST,  CouchbaseExceptionCodes::KEY_ALREADY_EXISTS])) {
        return FALSE;
      }
      throw $e;
    }
  }

  /**
   * Summary of touch
   *
   * @param string $id
   * @param int $expiry
   * @param array $options
   * @return mixed
   */
  public function touch($id, $expiry, array $options = []) {
    return $this->executeCouchbaseOperation(
      function() use ($id, $expiry, $options) {
        return $this->bucket->touch($id, $expiry, $options);
      }
    );
  }

  /**
   * Summary of counter
   *
   * @param mixed $ids
   * @param mixed $ids
   * @param array $options
   * @return mixed
   */
  public function counter($ids, $delta, array $options = []) {
    return $this->executeCouchbaseOperation(
      function() use ($ids, $delta, $options) {
        return $this->bucket->counter($ids, $delta, $options);
      }
    );
  }

  /**
   * Get the bucket manager.
   *
   * @return \CouchbaseBucketManager
   */
  public function manager() {
    return $this->bucket->manager();
  }

  /**
   * Escapes prefixes for cache strings identifiers.
   *
   * @param string $prefix
   *   Cache string prefix to escape.
   *
   * @return string
   *   Escaped prefix.
   */
  public function escapePrefix($prefix){
    $prefix = addcslashes($prefix, "%_\\\"");
    return addslashes($prefix);
  }

  /**
   * Delete elements by a prefix.
   *
   * @param string $prefix
   *   Prefix of documents to delete.
   *
   * @return bool
   *   Boolean indicating if deletion was succesful.
   */
  public function deleteAllByPrefix($prefix) {
    $prefix = $this->escapePrefix($prefix);

    $query = "DELETE FROM {$this->name} WHERE META({$this->name}).id LIKE \"{$prefix}%\"";
    $query = \CouchbaseN1qlQuery::fromString($query);

    $couchbase_manager = \Drupal\couchbasedrupal\CouchbaseManager::instance();
    if(($info = $this->manager()->info())
      && isset($info['name'])
      && isset($couchbase_manager->settings['servers'][$info['name']]['max_parallelism'])
    ){
      $query->max_parallelism = $couchbase_manager->settings['servers'][$info['name']]['max_parallelism'];
    }

    $result = $this->queryN1QL($query, NULL, TRUE);
    return isset($result->status) && $result->status === 'success';
  }

  /**
   * Delete elements by a prefix, or by an array of prefixes.
   *
   * @param string[] $prefixes
   */
  public function deleteAllByPrefixMultiple(array $prefixes) {
    $parts = [];
    $prefix_count = 0;
    $arguments = [];
    foreach ($prefixes as $prefix) {
      $parts[] = "META(self).id LIKE \prefix{$prefix_count}";
      $arguments["prefix{$prefix_count}"] = "{$this->escapePrefix($prefix)}%";
      $prefix_count++;
    }
    $query = \CouchbaseN1qlQuery::fromString("DELETE FROM {$this->name} WHERE " . implode(' AND ', $parts));
    $query->namedParams($arguments);

    $result = $this->queryN1QL($query, NULL, TRUE);
    return isset($result->status) && $result->status === 'success';
  }

  /**
   * Returns bucket name.
   *
   * @return string
   *   Bucket name.
   */
  public function getName() {
    return $this->name;
  }

  /**
   * Return all keys that start with a prefix
   * using N1QL.
   *
   * @param string $prefix
   * @return mixed
   */
  public function getAllKeysByPrefix($prefix) {
    $prefix = $this->escapePrefix($prefix);
    $prefix_length = strlen($prefix);

    $query = \CouchbaseN1qlQuery::fromString("SELECT SUBSTR(META({$this->name}).id, $prefix_length) AS cid, META({$this->name}).id as id FROM {$this->name} WHERE META({$this->name}).id LIKE '{$prefix}%'");
    $result = (array) $this->queryN1QL($query, NULL, TRUE);

    return array_combine(array_column($result, 'id'), array_column($result, 'cid'));
  }

  /**
   * Return all keys that start with a prefix
   * using N1QL.
   *
   * Implementation should not be based on document contents
   * because it is shared between Backend and RawBackend.
   *
   * @param string $prefix
   * @return mixed
   */
  public function getAllItemsByPrefix($prefix) {
    // Due to limitations in N1QL the most easy implementation
    // right now is to retrieve the keys and then the elements
    // because N1QL can only retrieve full documents and not
    // binary values.
    $keys = $this->getAllKeysByPrefix($prefix);
    return $this->getMultiple(array_keys($keys));
  }

  /**
   * Enables client side N1QL
   */
  public function enableN1ql() {
    if (!$this->N1qlEnabled) {
      $this->bucket->enableN1ql(['http://1.1.1.1:8093/','http://1.1.1.2:8093/']);
      $this->N1qlEnabled = TRUE;
    }
  }

  /**
   * https://forums.couchbase.com/t/how-to-execute-n1ql-queries-against-a-sasl-password-protected-bucket/7752
   * SASL auth buckets need the creds option to run N1QL queries.
   *
   * @param \CouchbaseN1qlQuery $query
   */
  protected function addCredentialsToQuery(\CouchbaseN1qlQuery &$query) {
    if (!empty($this->password)) {
      $query->options["creds"] = [
        ["user" => "local:{$this->name}", "pass" => $this->password]
      ];
    }
  }

  /**
   * Run a N1QL query
   *
   * @param string $query
   */
  public function queryN1QL(\CouchbaseN1qlQuery $query, $params = NULL, $json_array = FALSE) {
    $this->enableN1ql();
    try {
      $this->addCredentialsToQuery($query);
      return $this->executeCouchbaseOperation(
        function() use ($query, $params, $json_array) {
          // Basic instrumentation for N1Ql queries... watchdog
          // anything that runs beyond a threshold (in seconds).
          $threshold = 8;
          $start = microtime(TRUE);
          $result = $this->bucket->query($query, $params, $json_array);
          $duration = microtime(TRUE) - $start;
          if ($duration > $threshold && function_exists('watchdog')) {
            $options = (array) $query->toObject();
            watchdog('couchbasedrupal', 'Slow query run for :duration seconds :data', [':duration' => $duration, ':data' => json_encode($options)], WATCHDOG_DEBUG);
          }
          return $result;
        }
      );
    }
    catch (\CouchbaseException $exception){
      $this->handleQueryException($exception, $query, $this, $params);
    }
    catch (\Exception $e) {
      // CREATE INDEX idx ON mybucket(`key`, `value` IS VALUED) WHERE `value` IS VALUED;
      // https://forums.couchbase.com/t/explanation-on-primary-scan-secondary-scan/6048/12
      $code = isset($e->qCode) ? $e->qCode : $e->getCode();
      if ($code == 4000) {
        try {
          $fixquery = \CouchbaseN1qlQuery::fromString("CREATE PRIMARY INDEX ON {$this->name} USING GSI");
          $this->addCredentialsToQuery($fixquery);
          $this->bucket->query($fixquery);
        }
        catch (\CouchbaseException $exception){
          $this->handleQueryException($exception, $query, $this, $params);
        }
        return $this->bucket->query($query, $params, $json_array);
      }
      throw $e;
    }
  }

  /**
   * Wraps and re-throws any CouchbaseException exception.
   *
   * Bundles any couchbase exception thrown on CouchbaseBucket::queryN1QL with
   * it's query and query parameters to be processed by another higher exception.
   * This exception is passed to watchdog_exception to be logged for
   * development purpoises.
   *
   * @param \CouchbaseException $exception
   *   The exception thrown on CouchbaseBucket::queryN1QL.
   * @param $query
   *   The query to be executed.
   * @param \Drupal\couchbasedrupal\CouchbaseBucket $bucket
   *   The bucket that is processing the current query.
   * @param array $options
   *   An associative array of options to control how the query is run.
   */
  public function handleQueryException(\CouchbaseException $exception, $query, CouchbaseBucket $bucket, $options) {
    $exception->query_string = $query;

    $id = CouchbaseWatchdog::identifyException($exception);
    $message = "{$exception->getMessage()}<br>Exception identifier: [couchbasedrupal:{$id}]<br>";

    watchdog_exception('couchbasedrupal', $exception, $message, array('exception' => $exception));
  }

  /**
   * Run a view query
   *
   * @param \CouchbaseN1qlQuery $query
   * @param mixed $params
   * @param mixed $json_array
   * @return mixed
   */
  public function queryView(\CouchbaseViewQuery $query, $params = NULL, $json_array = FALSE) {
    if ($params == NULL && $json_array == NULL) {
      return $this->bucket->query($query);
    }
    else if ($json_array != FALSE) {
      return $this->bucket->query($query, $params, $json_array);
    }
    else {
      return $this->bucket->query($query, $params);
    }
  }

  /**
   * Clear all the documents (views) in this
   * bucket.
   */
  public function clearAllDocuments() {
    $manager = $this->manager();
    $documents = $manager->getDesignDocuments();
    foreach ($documents as $name => $definition) {
      $manager->removeDesignDocument($name);
    }
  }

  /**
   * Get information about an index, or false
   * if it does not exist.
   *
   * @param string $name
   * @return mixed
   */
  public function indexGet($name) {
    $result = $this->queryN1QL(\CouchbaseN1qlQuery::fromString("SELECT * FROM system:indexes WHERE name='$name'", NULL, FALSE));
    $result = json_decode(json_encode($result), true);
    if (isset($result['rows'][0]['indexes'])) {
      return $result['rows'][0]['indexes'];
    }
    return FALSE;
  }

  public function indexDrop($name) {
    $result = $this->queryN1QL(\CouchbaseN1qlQuery::fromString("DROP INDEX $name", NULL, FALSE));
  }
}
